#ifndef AMDIS_SECOND_ORDER_ASSEMBLER_H
#define AMDIS_SECOND_ORDER_ASSEMBLER_H

#include "SubAssembler.h"

namespace AMDiS {

  class Q11PsiPhi;

  /**
   * \ingroup Assembler
   * 
   * \brief
   * SubAssembler for second order terms.
   */
  class SecondOrderAssembler : public SubAssembler
  {
  public:
    MEMORY_MANAGED(SecondOrderAssembler);

    /** \brief
     * Creates and returns the SecondOrderAssembler for Operator op and
     * the given assembler. If all terms are piecewise constant precalculated 
     * integrals can be used while assembling and the returned 
     * ZeroOrderAssembler is of type Pre0. Otherwise a Quad0 object will
     * be returned.
     */
    static SecondOrderAssembler* getSubAssembler(Operator *op,
						 Assembler *assembler,
						 Quadrature *quadrat,
						 bool optimized);

    /** \brief
     * Destructor.
     */
    virtual ~SecondOrderAssembler() {};

  protected:
    /** \brief
     * Constructor.
     */
    SecondOrderAssembler(Operator *op,
			 Assembler *assembler,
			 Quadrature *quadrat,
			 bool optimized);

  protected:
    /** \brief
     * List of all yet created optimized second order assemblers.
     */
    static std::vector<SubAssembler*> optimizedSubAssemblers;

    /** \brief
     * List of all yet created standard second order assemblers.
     */
    static std::vector<SubAssembler*> standardSubAssemblers;
  };

  // ============================================================================
  // ===== class Stand2 =========================================================
  // ============================================================================

  /**
   * \ingroup Assembler
   * 
   * \brief
   * Standard second order assembler
   */
  class Stand2 : public SecondOrderAssembler 
  {
  public:
    MEMORY_MANAGED(Stand2);

    /** \brief
     * Constructor.
     */
    Stand2(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    void calculateElementMatrix(const ElInfo *rowElInfo,
				const ElInfo *colElInfo,
				const ElInfo *smallElInfo,
				const ElInfo *largeElInfo,
				ElementMatrix *mat) 
    {
      ERROR_EXIT("CEM not yet\n");
    }

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector */*vec*/) {
      ERROR_EXIT("should not be called\n");
    };
  };

  // ============================================================================
  // ===== class Quad2 ==========================================================
  // ============================================================================

  /** \brief
   * Second order assembler using fast quadratures.
   */
  class Quad2 : public SecondOrderAssembler 
  {
  public:
    MEMORY_MANAGED(Quad2);

    /** \brief
     * Constructor.
     */
    Quad2(Operator *op, Assembler *assembler, Quadrature *quad);

    ~Quad2();

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    void calculateElementMatrix(const ElInfo *rowElInfo,
				const ElInfo *colElInfo,
				const ElInfo *smallElInfo,
				const ElInfo *largeElInfo,
				ElementMatrix *mat) 
    {
      ERROR_EXIT("CEM not yet\n");
    }

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector */*vec*/) {
      ERROR_EXIT("should not be called\n");
    };

  protected:
    /** \brief
     * Thread safe temporary vector of DimMats for calculation in calculateElementMatrix().
     */
    std::vector< DimMat<double>** > tmpLALt;
  };

  // ============================================================================
  // ===== class Pre2 ===========================================================
  // ============================================================================

  /** \brief
   * Second order assembler using predefined integrals.
   */
  class Pre2 : public SecondOrderAssembler 
  {
  public:
    MEMORY_MANAGED(Pre2);

    /** \brief
     * Constructor.
     */
    Pre2(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Destructor.
     */
    ~Pre2();

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    void calculateElementMatrix(const ElInfo *rowElInfo,
				const ElInfo *colElInfo,
				const ElInfo *smallElInfo,
				const ElInfo *largeElInfo,
				ElementMatrix *mat) 
    {
      ERROR_EXIT("CEM not yet\n");
    }

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector *) {
      ERROR_EXIT("should not be called\n");
    };

  protected:
    /** \brief
     * Integral of the product of the derivative of psi and the derivative 
     * of phi.
     */
    const Q11PsiPhi *q11;

    /** \brief
     * Thread safe temporary vector of DimMats for calculation in calculateElementMatrix().
     */
    std::vector< DimMat<double>** > tmpLALt;

    friend class SecondOrderAssembler;
  };

}

#endif // AMDIS_SECOND_ORDER_ASSEMBLER_H
